#!/bin/sh
#
# Verify external storage medium and mount it
#
# Note: It is required that the external storage is not automatically
# mounted by the mountall boot step (fstab: noauto).

FORMAT=0
sdoptions=''
minpartsize=$((128 * 1024))
REBOOT_LOG=/var/log/sdcard/reboots.log
FSCK_LOG=/var/log/sdcard/fsck.log

# read by dss11-websetup/www/cgi-bin/libsdcard.sh
FSCK_STATUS=/var/run/sdcard-fsck.status

isnum() {
    # http://stackoverflow.com/questions/2210349/bash-test-whether-string-is-valid-as-an-integer
    [ -n "$1" -a $1 -eq $1 ] 2> /dev/null
}

request_reboot()
{
    # ensure file exists and we have permissions
    # if we err' here we can create a nasty reboot loop
    touch $REBOOT_LOG
    err=$?
    if [ $err -ne 0 ]; then
        echo "Can't access $REBOOT_LOG, ignore reboot request"
        return 250
    fi

    ct=`wc -l $REBOOT_LOG | cut -f1 -d\ `
    echo "dbg> ct: \"$ct\""

    if isnum $ct; then
        echo "Failed to count lines of $REBOOT_LOG", ingore reboot request
        exit 249
    fi

    echo "number of reboots: $ct"
    if [ $ct -gt 5 ]; then
        echo "Too many reboots requested, continue without card"
        return 255
    fi

    echo "schedule reboot"
    date +"%Y-%m-%dT%H:%M:%S" >> $REBOOT_LOG
    shutdown -r now
}

format_sd()
{
    echo "SD-CARD creating ext2 filesystem..."
    echo 1 > /sys/class/leds/blue/brightness
    echo 0 > /sys/class/leds/green/brightness
    echo 1 > /sys/class/leds/red/brightness

    exec 7>&1
    exec 8>&2
    exec >> $FSCK_LOG 2>&1
    mkfs.ext2 -v -L DSSDATA $sdoptions /dev/mmcblk0p1
    error=$?
    exec 1>&7 7>&-
    exec 2>&8 8>&-
    echo $error > /var/sdcard-mkfs.status # dss11-websetup transition
    echo $error > $FSCK_STATUS
    return $error
}

fsck_sd()
{
    mode=$1

    exec 7>&1
    exec 8>&2
    exec >> $FSCK_LOG 2>&1
    if [ "x$mode" = "xautorepair" ]; then
        echo "[autorepair]"
        fsck.ext4 -v -y -f /dev/mmcblk0p1
        error=$?
        echo $error > /tmp/sdcard-fsck.status # dss11-websetup transition
        echo $error > $FSCK_STATUS
    else
        fsck.ext4 -v -p /dev/mmcblk0p1
        error=$?
        echo $error >/tmp/sdcard-fsck.status # dss11-websetup transition
        echo $error > $FSCK_STATUS
    fi
    exec 1>&7 7>&-
    exec 2>&8 8>&-

    echo -n "SD-CARD Status: "
    cat $FSCK_STATUS
    return $error
}

mount_sd()
{
    # in fstab it's noauto, but better safe than sorry
    umount /dev/mmcblk0p1 > /dev/null 2>&1

    if [ ! -b /dev/mmcblk0 ]; then
        echo "SD-CARD not found, aborting... "
        request_reboot
        return 254
    fi

    # fdisk will trigger update of /sys/block/mmcblk0
    if [ ! -b /dev/mmcblk0p1 ]; then
        echo "SD-CARD invalid partition table: re-create..."
        # new partition, primary, partition 1, default, default, w
        echo "n
        p
        1


        w
        "| fdisk /dev/mmcblk0
        FORMAT=1
    fi

    if [ ! -d /sys/block/mmcblk0/mmcblk0p1 ]; then
        echo "SD-CARD meta information not available, aborting..."
        request_reboot
        return 253
    fi

    sdsize=$(cat /sys/block/mmcblk0/mmcblk0p1/size)
    eval `blkid /dev/mmcblk0p1 | cut -f2 -d:`
    if ((test -n "$TYPE") &&
        (test "$TYPE" != "ext4") &&
        (test "$TYPE" != "ext3") &&
        (test "$TYPE" != "ext2")) ; then
        echo "SD-CARD filesystem \'$TYPE\' is unusable and will be reformatted"
        FORMAT=1
    fi

    if [ $sdsize -le $minpartsize ]; then
        # TODO maybe card is partitioned wrong, check size of card
        echo "SD-CARD partition too small, ignore.... "
        return 255
    fi

    if [ -f /etc/.sdformat ]; then
        FORMAT=1
        sdoptions=`head -n 1 /etc/.sdformat` # bogus, see cgi-bin/format.sh
        echo "SD-CARD requested reformat with options: \"$sdoptions\""
        # mkfs.ext4 options, eg. "-cc" for intensive bad block tests
    fi

    if [ $FORMAT -eq 0 ]; then
        echo "SD-CARD filesystem check..."
        fsck_sd
        error=$?

        if [ $(($error & 4)) == 4 ] ; then
            echo "SD-CARD automatic repair filesystem"
            fsck_sd autorepair
            error=$?
        fi

        if [ $(($error & 2)) == 2 ] ; then
            echo "SD-CARD fsck requested reboot"
            request_reboot
            return 126
        fi

        if [ $(($error & ~3)) != 0 ] ; then
            echo "SD-CARD: Fatal errors on SD-CARD detected [fsck:$error]"
            FORMAT=1
        fi
    fi

    # aggregate .sdformat/re-partition/blkid format requests
    if [ $FORMAT -eq 1 ]; then
        format_sd
        if [ $? -ne 0 ]; then
            echo "SD-CARD: WARNING failed to format SD-CARD"
            request_reboot
            return 125
        fi

        # paranoia
        fsck_sd
        error=$?
        if [ $error -ne 0 ]; then
            echo "SD-CARD: WARNING new filesystem has errors"
            request_reboot
            return 124
        fi
        rm -f /etc/.sdformat
    fi

    # journal increases number of writes, disable
    dumpe2fs -h /dev/mmcblk0p1 | grep -e "^Filesystem\ features:[ \t]*has_journal"
    ret=$?
    if [ $ret -eq 0 ]; then
        echo "Disable journal on sdcard"
        tune2fs -O ^has_journal /dev/mmcblk0p1
    fi

    mount /media/card

    # fresh start
    echo -n > $REBOOT_LOG
}

prepare_overlay()
{
    #
    #  /etc/fstab is maintained by base-files package
    #  we need to create the branch directory for those
    #  mounts to work
    #
    #  mkdir -p /media/card/sys/log
    #  mkdir -p /media/card/sys/metering
    #
    echo "SD-CARD creating filesystem overlays"
    while read DIR; do
        echo "  $DIR"
        mkdir -p $DIR
    done <<EOF
    $(grep ^aufs /etc/fstab | sed -n 's/.*,br:\([^,=]*\).*/\1/p')
EOF
}

streamcopy ()
{
    source=$1
    dest=$2
    options=$3
    [ -d $dest ] || mkdir -p $dest
    (cd $source; tar -cf - --exclude=".*" $options * | tar -C $dest -xf -)
}

move2internal ()
{
    echo -n "SD-CARD merging data to internal storage: "

    #
    # transitional step:
    # previously, most user data was stored on sd-card
    # move those data back into more reliable internal
    # flash. executed only once per installation
    #

    if [ -e /media/card/ssl ]; then
        echo -n "Certificates "
        streamcopy /media/card/ssl/ /etc/ssl/certs/
        mv /media/card/ssl /media/card/ssl.moved
    fi

    if [ -e /media/card/sys/dss/data/metering ]; then
        echo -n "Metering "
        streamcopy /media/card/sys/dss/data/metering/ /media/card/sys/metering/
        rm -rf /media/card/sys/dss/data/metering
    fi

    if [ -e /media/card/sys/dss/data ]; then
        echo -n "dSS-Runtime "
        streamcopy /media/card/sys/dss/data/ /usr/share/dss/data/
        mv /media/card/sys/dss/data /media/card/sys/dss/data.moved
    fi

    if [ -e /media/card/sys/dss/add-ons ]; then
        echo -n "dSS-Addons "
        streamcopy /media/card/sys/dss/add-ons/ /usr/share/dss/add-ons/
        mv /media/card/sys/dss/add-ons /media/card/sys/dss/add-ons.moved
    fi

    if [ -e /media/card/sys/add-ons ]; then
        echo -n "Addons "
        streamcopy /media/card/sys/add-ons/ /www/pages/add-ons/
        mv /media/card/sys/add-ons /media/card/sys/add-ons.moved
    fi

    if [ -e /media/card/sys/dsa ]; then
        echo -n "dSA-Database (ignored)"
        #streamcopy /media/card/sys/dsa/ /var/lib/dsa
        #mv /media/card/sys/dsa /media/card/sys/dsa.moved
    fi

    echo ""
    sync
}

if [ -n "$UNITTEST" ]; then
    # exit will abort the unit test
    return
fi

#
# MAIN
#

#
# The opkg install hook runs while the sd-card is mounted. That means
# /var/log/sdcard folder is created on the sd-card. But the sd-card is
# not yet mounted. Ensure that folder exists or hell breaks loose
#
mkdir -p /var/log/sdcard

#
# will partition, format, mount the sd card.
# incl. retries and reboots
#
mount_sd

if mountpoint -q /media/card; then
    echo "SD-CARD: using sd-card LABEL:$LABEL UUID:$UUID "
    move2internal

    # check and prepare overlay mounts
    prepare_overlay
    mount -a -t aufs || error=$?
fi
